@@ -20,7 +20,11 @@ import stat
 import subprocess
 import sys
 
-from gi.repository import GLib as glib
+# Try to import the dynamic glib, or try to fall back to static
+try:
+    from gi.repository import GLib as glib  # pragma: no cover
+except ImportError:                         # pragma: no cover
+    import glib                             # pragma: no cover
 
 import dbus
 import dbus.mainloop.glib
@@ -40,7 +44,7 @@ def resolve_path(cmdname):
 
 # Constants
 NETWORKCTL = resolve_path('networkctl')
-DEFAULT_SCRIPT_DIR = '/etc/networkd-dispatcher'
+DEFAULT_SCRIPT_DIR = '/etc/networkd-dispatcher:/usr/lib/networkd-dispatcher'
 
 # Supported wireless tools
 IWCONFIG = resolve_path('iwconfig')
@@ -48,7 +52,6 @@ IW = resolve_path('iw')
 
 LOG_FORMAT = '%(levelname)s:%(message)s'
 
-STATE_IGN = {'carrier', 'degraded', None}
 SINGLETONS = {'Type', 'ESSID', 'OperationalState'}
 
 
@@ -74,7 +77,7 @@ def get_networkctl_list():
     """Update the mapping from interface index numbers to state"""
     try:
         out = subprocess.check_output([NETWORKCTL, 'list', '--no-pager',
-                                      '--no-legend'])
+                                       '--no-legend'])
     except subprocess.CalledProcessError as e:
         logger.error('networkctl list failed: %s', e)
         return []
@@ -93,7 +96,7 @@ def get_networkctl_status(iface_name):
     data = collections.defaultdict(list)
     try:
         out = subprocess.check_output([NETWORKCTL, 'status', '--no-pager',
-                                      '--no-legend', '--', iface_name])
+                                       '--no-legend', '--', iface_name])
     except subprocess.CalledProcessError as e:
         logger.error('Failed to get interface "%s" status: %s', iface_name, e)
         return data
@@ -116,7 +119,7 @@ def get_wlan_essid(iface_name):
     if IWCONFIG is None:
         if IW is None:
             logger.error('Unable to retrieve ESSID for wireless interface %s: '
-                         'no supported wireless tool installed' % iface_name)
+                         'no supported wireless tool installed', iface_name)
             return ''
         return iw_get_ssid(iface_name)
     return iwconfig_get_ssid(iface_name)
@@ -126,9 +129,9 @@ def iw_get_ssid(iface_name):
     out = subprocess.check_output([IW, iface_name, 'link'])
     lines = out.decode('utf-8', errors='replace').split('\n')
     line = [s for s in lines if 'SSID' in s]
-    if len(line) == 0:
-        logger.warning('Unable to retrieve ESSID for wireless interface %s.'
-                       % iface_name)
+    if not line:
+        logger.warning('Unable to retrieve ESSID for wireless interface %s.',
+                       iface_name)
         return ''
     essid = line[0].rsplit(" ")[1]
     return unquote(essid)
@@ -141,28 +144,41 @@ def iwconfig_get_ssid(iface_name):
     return unquote(essid)
 
 
-def scripts_in_path(path):
-    """Given a directory name, return a sorted list of executables
-    contained therein"""
+def scripts_in_path(path, subdir):
+    """Given directory names in PATH notation (separated by :), and a
+    subdirectory name, return a sorted list of executables
+    contained in that subdirectory, such that executables in earlier
+    path components override those with the same name in later path
+    components."""
     script_list = []
-    if not os.path.exists(path):
-        logger.debug("Path %r does not exist; skipping", path)
-        return []
-    for filename in os.listdir(path):
-        pathname = os.path.join(path, filename)
-        if os.path.isfile(pathname):
-            entry = os.stat(pathname)
-            # Make sure script can be executed
-            if not stat.S_IXUSR & entry.st_mode:
-                logger.error("Unable to execute script, check file mode: %s",
-                             pathname)
-            # Make sure script is owned by root
-            elif entry.st_uid != 0 or entry.st_gid != 0:
-                logger.error("Unable to execute script, check file perms: %s",
-                             pathname)
-            else:
-                script_list.append(pathname)
-    return sorted(script_list)
+    base_filenames = set()
+    for one_path in path.split(":"):
+        one_path = os.path.join(one_path, subdir)
+        if not os.path.exists(one_path):
+            logger.debug("Path %r does not exist; skipping", one_path)
+            continue
+        base_filenames.update(os.listdir(one_path))
+
+    for filename in sorted(base_filenames):
+        for one_path in path.split(":"):
+            pathname = os.path.join(one_path, subdir, filename)
+            logger.debug("Checking if %s exists as %s", filename, pathname)
+
+            if os.path.isfile(pathname):
+                entry = os.stat(pathname)
+                # Make sure script can be executed
+                if not stat.S_IXUSR & entry.st_mode:
+                    logger.error("Unable to execute script, check file mode: "
+                                 "%s", pathname)
+                # Make sure script is owned by root
+                elif entry.st_uid != 0 or entry.st_gid != 0:
+                    logger.error("Unable to execute script, check file perms: "
+                                 "%s", pathname)
+                else:
+                    script_list.append(pathname)
+                break
+
+    return script_list
 
 
 def parse_address_strings(addrs):
@@ -181,11 +197,12 @@ def parse_address_strings(addrs):
     return AddressList(ip4addrs, ip6addrs)
 
 
-def get_interface_data(iface, state):
+def get_interface_data(iface):
     """Return JSON-serializable data representing all state needed to run
     hooks for the given interface"""
     data = {'Type': iface.type, 'OperationalState': iface.operational,
-            'AdministrativeState': iface.administrative}
+            'AdministrativeState': iface.administrative,
+            "InterfaceName": iface.name}
     # Always collect what data we can.
     data.update(get_networkctl_status(iface.name))
     # The returned state may be different than what was read from
@@ -198,10 +215,11 @@ def get_interface_data(iface, state):
     return data
 
 
-class Dispatcher(object):
+class Dispatcher():
+    iface_names_by_idx = {}    # only changed on rescan
+    ifaces_by_name = {}        # updated on every state change
+
     def __init__(self, script_dir=DEFAULT_SCRIPT_DIR):
-        self.iface_names_by_idx = {}    # only changed on rescan
-        self.ifaces_by_name = {}        # updated on every state change
         self.script_dir = script_dir
         self._interface_scan()
 
@@ -246,7 +264,7 @@ class Dispatcher(object):
 
     def get_scripts_list(self, state):
         """Return scripts for the given state"""
-        return scripts_in_path(self.script_dir + "/" + state + ".d")
+        return scripts_in_path(self.script_dir, state + ".d")
 
     def _handle_one_state(self, iface_name, state, state_type, force=False):
         """Process a single state change"""
@@ -257,7 +275,7 @@ class Dispatcher(object):
             prior_iface = self.ifaces_by_name.get(iface_name)
             if prior_iface is None:
                 logger.error('Attempting to handle state for unknown interface'
-                             '%r', iface_name)
+                             ' %r', iface_name)
                 return
 
             prior_state = getattr(prior_iface, state_type)
@@ -269,11 +287,6 @@ class Dispatcher(object):
             new_iface = prior_iface._replace(**{state_type: state})
             self.ifaces_by_name[new_iface.name] = new_iface
 
-            if state in STATE_IGN:
-                logger.debug('Ignored state %r seen for interface %r, '
-                             'skipping', state, iface_name)
-                return
-
             self.run_hooks_for_state(new_iface, state)
         # pylint: disable=broad-except
         except Exception:
@@ -298,7 +311,7 @@ class Dispatcher(object):
             return
 
         # Collect data
-        data = get_interface_data(iface, state)
+        data = get_interface_data(iface)
         (v4addrs, v6addrs) = parse_address_strings(data.get('Address', ()))
 
         # Set script env. variables
@@ -312,7 +325,7 @@ class Dispatcher(object):
             'STATE': str(state),
             'AdministrativeState': data.get('AdministrativeState', ''),
             'OperationalState': data.get('OperationalState', ''),
-            'json': json.dumps(data),
+            'json': json.dumps(data, sort_keys=True),
         })
 
         # run all valid scripts in the list
@@ -373,11 +386,9 @@ class Dispatcher(object):
                 self.ifaces_by_name.pop(iface_name)
             except KeyError:
                 logger.error('Unable to remove interface at index %r.', idx)
-            finally:
-                return
 
 
-def sd_notify(unset_environment=False, **kwargs):
+def sd_notify(**kwargs):
     """Systemd sd_notify implementation for Python.
     Note: kwargs should contain the state to send to systemd"""
     if not kwargs:
@@ -404,19 +415,17 @@ def sd_notify(unset_environment=False, **kwargs):
             return 1
     # pylint: disable=broad-except
     except Exception:
-        logger.exception("Ignoring unexpected error during sd_notify()"
+        logger.exception("Ignoring unexpected error during sd_notify() "
                          "invocation")
 
     if sock:
         sock.close()
-    if unset_environment:
-        if 'NOTIFY_SOCKET' in os.environ:
-            del os.environ['NOTIFY_SOCKET']
 
     return 0
 
 
-def main():
+def parse_args(args):
+    """Parses arguments from command line"""
     ap = argparse.ArgumentParser(description='networkd dispatcher daemon')
     ap.add_argument('-S', '--script-dir', action='store',
                     default=DEFAULT_SCRIPT_DIR,
@@ -429,7 +438,11 @@ def main():
                     help='Increment verbosity level once per call')
     ap.add_argument('-q', '--quiet', action='count', default=0,
                     help='Decrement verbosity level once per call')
-    args = ap.parse_args()
+    return ap.parse_args(args)
+
+
+def main():
+    args = parse_args(sys.argv[1:])
 
     verbosity_num = (args.verbose - args.quiet)
     if verbosity_num <= -2:
@@ -466,7 +479,10 @@ def main():
     mainloop.run()
 
 
-if __name__ == '__main__':
-    main()
+def init():
+    if __name__ == '__main__':
+        main()
 
+
+init()
 # vim: ai et sts=4 sw=4 ts=4
